// --- [ Master_Slave.c ] -----------------------------------------------------
//       10.11.08		tab = 3
//       auf der Grundlage der AppNote AVR 311 / 315 von Atmel
//       fr TWI-Master und TWI-Slave zusammen
//       Master und Slave haben getrennte Buffer
//       Fr den Master ist ein Repeated Start mglich , indem eine
//       Transaktion NICHT mit einem STOP beendet wird (-> TWI_MA_no_stop)
//       Beim Empfangen wird als 1. Byte die vom Master gesendete Adresse
//       (das ist die eigene Adresse !) bergeben.
//
//       Fehlerbehandlung:
//       TWSR wird nach TWI_state kopiert, wenn in der TWI_ISR nicht behandelter Status 
//       vorgefunden wird. Ohne Fehler ist TWI_state == TWI_NO_STATE.
//       Eine fehlerfreie bertragung wird auch durch TWI_StatusReg.LastTransOK signalisiert.
//
//       Problem
//       TWI_Transceiver_Busy() erkennt am Flag TWIE, dass eine Transaktion im Gange ist.
//       Im Master-Modus funktioniert das auch.      
//       Allerdings ist dieses Flag whrend des gesamten Slave-Modus gesetzt:
//       Der Aufruf von TWI_Transceiver_Busy() im Slave-Modus fhrt in eine Endlossschleife.
//       -> daher Einfhrung von TWI_Mode und TWI_Slave_Busy() / TWI_Master_Busy()
// ----------------------------------------------------------------------------


#include <avr/interrupt.h>              
#include <avr/io.h>
#include <inttypes.h>
#include "TWI_Master_Slave.h"
 
static uint8_t TWI_SLA_buf[SLA_BUFFER_SIZE]; // Die Puffer fr den Master und den Slave sind getrennt
static uint8_t TWI_MA_buf[MA_BUFFER_SIZE];   // Nach Arbitrierungsfehler bleiben die Sendedaten unverndert erhalten
															// und knnen mit mit ein TWI_MA_Start_Transceiver() neu gesendet werden
static uint8_t TWI_Arbitration = 0;				// Flag, das bei Arbitrierungsfehler mit Inhalt von TSWR beschrieben wird
static uint8_t TWI_msgSize  = 0;           	// Number of bytes to be transmitted.
static uint8_t TWI_state    = TWI_NO_STATE;	// State byte. Default set to TWI_NO_STATE.

volatile uint8_t TWI_SLA_bufPtr = 0;			// enthlt die Anzahl der vom Slave empfangenen Bytes
volatile uint8_t TWI_Mode;

volatile uint8_t TWI_MA_no_stop = FALSE;		// TRUE signalisiert, dass kein STOP gesendet werden soll (repeated start)

union TWI_statusReg TWI_statusReg = {0};     // TWI_statusReg is defined in TWI_Slave.h

uint8_t TWI_Transceiver_Busy(void);
uint8_t TWI_Slave_Busy(void);
uint8_t TWI_Master_Busy(void);

// anstelle des Exportes mit einer public-Variablen wird TWI_statusReg als lokale Variable ber Funktionen abgefragt
// ----------------------------------------------------------------------------
uint8_t TWI_LastTransOK(void)
{
	return(TWI_statusReg.LastTransOK);
}
/*
// ----------------------------------------------------------------------------
uint8_t TWI_RxDataInBuf(void)				      // besser:  TWI_SLA_RxCount()
{
return(TWI_statusReg.RxDataInBuf);
}
*/

// ----------------------------------------------------------------------------
uint8_t TWI_GenAddressCall(void)
{
	return(TWI_statusReg.GenAddressCall);
}

// ----------------------------------------------------------------------------
// liefert 0, wenn kein Arbitrierungsfehler vorliegt - und sonst den TWSR-Code
uint8_t TWI_Get_Arbitration(void)
{
	return(TWI_Arbitration); 
}

// ----------------------------------------------------------------------------
// Grundlegende Initailisierung fr Master und Slave Modus
// Den Unterschied zwischen Master und Slave macht lediglich das Bit TWEA
// (damit antwortet der Slave mit einem ACK, sobald er addressiert wird)
// Fr den Master Modus mssen TWIE / TWINT gesetzt werden
// Fr den Slave Modus  muss zustzlich TWEA gesetzt werden
void TWI_Initialise( uint8_t TWI_ownAddress )
{
	TWI_Mode = 0;
   TWBR = TWI_TWBR;                             // Set bit rate register (Baudrate). Defined in header file.
// TWSR = TWI_TWPS;                             // Not used. Driver presumes prescaler to be 00.
	TWAR = TWI_ownAddress;                       // Set own TWI slave address.
	TWDR = 0xFF;                                 // Default content = SDA released.
	TWCR = (1<<TWEN)|                          	// Enable TWI-interface and release TWI pins.
          (0<<TWIE)|(0<<TWINT)|                 // Disable TWI Interupt.
          (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|      // Do not ACK on any requests, yet.
          (0<<TWWC);                            
//	PORT_TWI |= (1<<PIN_SCL | 1<<PIN_SDA);			// SDA und SCL auf high
//	DDR_TWI  |= (1<<PIN_SDA | 1<<PIN_SCL);			// SDA und SCL als Ausgang, wenn nicht das TWI-Modul aktiv ist
}    

   
// ----------------------------------------------------------------------------
// liefert .T., wenn der Master ODER der Slave aktiv ist
// liefert .F., wenn keiner von beiden aktiv
uint8_t TWI_Transceiver_Busy( void )
{
	uint8_t i = 0;
	i  = ((TWI_Mode & (1<<MASTER | 1<<ACTIV)) == (1<<MASTER | 1<<ACTIV));
	i += ((TWI_Mode & (1<<SLAVE  | 1<<ACTIV)) == (1<<SLAVE  | 1<<ACTIV));
//	i  = TWI_Master_Busy();
//	i += TWI_Slave_Busy();

	return (i); 
} 

// ----------------------------------------------------------------------------
// liefert .T. wenn der Master aktiv ist

uint8_t TWI_Master_Busy(void)
{
	return ((TWI_Mode & (1<<MASTER | 1<<ACTIV)) == (1<<MASTER | 1<<ACTIV));
}

// ----------------------------------------------------------------------------
// liefert .T. wenn der Slave aktiv ist

uint8_t TWI_Slave_Busy(void)
{
	return ((TWI_Mode & (1<<SLAVE | 1<<ACTIV)) == (1<<SLAVE | 1<<ACTIV));
}

/****************************************************************************
Call this function to fetch the state information of the previous operation. 
The function will hold execution (loop) until the TWI_ISR has completed with the previous operation. 
If there was an error, then the function will return the TWI State code. 
****************************************************************************/
/*uint8_t TWI_Get_State_Info( void )
{
	while (( TWI_Master_Busy()) || (TWI_Slave_Busy()) ); // Wait until TWI has completed the transmission.
	return ( TWI_state );                         // Return error state. 
} */

// ----------------------------------------------------------------------------
uint8_t TWI_Get_Err(void)
{													 // Wait until TWI has completed the transmission.
	while ((TWI_Master_Busy()) || TWI_Slave_Busy());
//	while (TWI_Transceiver_Busy());
	if (TWI_state == TWI_NO_STATE) return(0);
	return(TWI_state);
}


/****************************************************************************
Call this function to send a prepared message. The first byte must contain the slave address and the
read/write bit. Consecutive bytes contain the data to be sent, or empty locations for data to be read
from the slave. Also include how many bytes that should be sent/read including the address byte.
The function will hold execution (loop) until the TWI_ISR has completed with the previous operation,
then initialize the next operation and return.
****************************************************************************/
void TWI_MA_Start_Transceiver_With_Data( uint8_t *msg, uint8_t msgSize )
{
   uint8_t temp;

   while ( TWI_Transceiver_Busy() );           		// Wait until Master is ready for next transmission.

	TWI_Arbitration   = 0;								// Flag nur fr den Multimaster-Betrieb
   TWI_statusReg.all = 0;
   TWI_state         = TWI_NO_STATE ;
	
   TWI_msgSize = msgSize;                      	// Number of data to transmit.
   TWI_MA_buf[0]  = msg[0];                    	// Store slave address with R/W setting.
   if (!( msg[0] & 0x01))						     	// If it is a write operation, then also copy data.
		{
		for ( temp = 1; temp < msgSize; temp++ )
			TWI_MA_buf[ temp ] = msg[ temp ];
		}
	TWI_Mode = (1<<MASTER | 1<< ACTIV | 1<<TRANSMIT);
   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|      // Initiate a START condition.
          (0<<TWWC);     
}

/****************************************************************************
Call this function to resend the last message. The driver will reuse the data previously put in the transceiver buffers.
The function will hold execution (loop) until the TWI_ISR has completed with the previous operation,
then initialize the next operation and return.
****************************************************************************/
void TWI_MA_Start_Transceiver( void )
{
   while ( TWI_Transceiver_Busy() );            // Wait until TWI is ready for next transmission.
	TWI_Arbitration 	= 0;								// Flag nur fr den Multimaster-Betrieb
   TWI_statusReg.all = 0;      
   TWI_state         = TWI_NO_STATE ;
	
	TWI_Mode = (1<<MASTER | 1<< ACTIV | 1<<TRANSMIT);
   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|      // Initiate a START condition.
          (0<<TWWC);                           
}


/****************************************************************************
Call this function to read out the requested data from the TWI transceiver buffer. I.e. first call
TWI_Start_Transceiver to send a request for data to the slave. Then Run this function to collect the
data when they have arrived. Include a pointer to where to place the data and the number of bytes
requested (including the address field) in the function call. The function will hold execution (loop)
until the TWI_ISR has completed with the previous operation, before reading out the data and returning.
If there was an error in the previous transmission the function will return the TWI error code.
****************************************************************************/
uint8_t TWI_MA_Get_Data_From_Transceiver( uint8_t *msg, uint8_t msgSize )
{
   uint8_t i;

   while ( TWI_Master_Busy() );            		// Wait until TWI is ready for next transmission.

   if( TWI_statusReg.LastTransOK )              // Last transmission competed successfully.
		{                                             
		for ( i=0; i<msgSize; i++ )               // Copy data from Transceiver buffer.
			{
			msg[ i ] = TWI_MA_buf[ i ];
			}
		}
   return( TWI_statusReg.LastTransOK );
}


/****************************************************************************
Call this function to send a prepared message, or start the Transceiver for reception. 
Include a pointer to the data to be sent if a SLA+W is received. The data will be copied to the TWI buffer. 
Also include how many bytes that should be sent. 
Note that unlike the similar Master function, the Address byte is not included in the message buffers.
The function will hold execution (loop) until the TWI_ISR has completed with the previous operation,
then initialize the next operation and return.
****************************************************************************/
void TWI_SLA_Start_Transceiver_With_Data( uint8_t *msg, uint8_t msgSize )
{
	uint8_t temp;

	while ( TWI_Transceiver_Busy() );             // Wait until TWI is ready for next transmission.

	TWI_msgSize = msgSize;                        // Number of data to transmit.
	for ( temp = 0; temp < msgSize; temp++ )      // Copy data that may be transmitted if the TWI Master requests data.
		TWI_SLA_buf[ temp ] = msg[ temp ];
	TWI_statusReg.all = 0;      
	TWI_state         = TWI_NO_STATE ;
	TWI_Mode = (1<<SLAVE);
	TWCR = (1<<TWEN)|                             // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                  // Enable TWI Interupt and clear the flag.
          (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|       // Prepare to ACK next time the Slave is addressed.
          (0<<TWWC);                             //
}

/****************************************************************************
Call this function to start the Transceiver without specifing new transmission data. 
Usefull for restarting a transmission, or just starting the transceiver for reception. 
The driver will reuse the data previously put in the transceiver buffers. 
The function will hold execution (loop) until the TWI_ISR has completed with the 
previous operation, then initialize the next operation and return.
****************************************************************************/
void TWI_SLA_Start_Transceiver( void )
{
	while ( TWI_Transceiver_Busy() );             // Wait until TWI is ready for next transmission.
	TWI_statusReg.all = 0;      
	TWI_state         = TWI_NO_STATE ;
	TWI_Mode = (1<<SLAVE);
	TWCR = (1<<TWEN)|                             // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                  // Enable TWI Interupt and clear the flag.
          (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|       // Prepare to ACK next time the Slave is addressed.
          (0<<TWWC);                             //
}

/****************************************************************************
Liefert die Anzahl der im Slave-Modus empfangenen Bytes zurck
Solange eine SLA-bertragung nicht abgeschlossen ist, wird 0 zurckgeliefert.
****************************************************************************/
uint8_t TWI_SLA_RxCount(void)
{
	// keine Werte zurckgeben, solange wie der Slave aktiv ist 
	if (TWI_Slave_Busy() ) return(0);	

return(TWI_SLA_bufPtr);
}

/****************************************************************************
Call this function to read out the received data from the TWI transceiver buffer. 
I.e. first call TWI_Start_Transceiver to get the TWI Transceiver to fetch data. 
Then Run this function to collect the data when they have arrived. 
Include a pointer to where to place the data and the number of bytes to fetch in the function call. 
The function will hold execution (loop) until the TWI_ISR has completed with the previous operation, 
before reading out the data and returning.
If there was an error in the previous transmission the function will return the TWI State code.
****************************************************************************/
uint8_t TWI_SLA_Get_Data_From_Transceiver( uint8_t *msg, uint8_t msgSize )
{
	uint8_t i;

	while ( TWI_Slave_Busy() );             		// Wait until TWI is ready for next transmission.

	if( TWI_statusReg.LastTransOK )              // Last transmission completed successfully.              
		{                                             
		for ( i=0; i<msgSize; i++ )               // Copy data from Transceiver buffer.
			{
			msg[ i ] = TWI_SLA_buf[ i ];
			}
		TWI_statusReg.RxDataInBuf = FALSE;        // Slave Receive data has been read from buffer.
		}
	TWI_SLA_bufPtr = 0;
	return( TWI_statusReg.LastTransOK );                                   
}

/****************************************************************************
This function is the Interrupt Service Routine (ISR), and called when the TWI interrupt is triggered;
that is whenever a TWI event has occurred. This function should not be called directly from the main
application.
****************************************************************************/

ISR(TWI_vect)
{
static uint8_t TWI_bufPtr;

switch (TWSR)
	{
// --- SLAVE ------------------------------------------------------------------
// --- TWI Slave Transmitter status codes -------------------------------------
																
	case TWI_STX_ADR_ACK_M_ARB_LOST: 				// Arbitration lost in SLA+R/W as Master; 
																// own SLA+R has been received; ACK has been returned
		TWI_Arbitration 	= TWSR;						// Flag fr Arbitrierungsfehler setzen
		
   case TWI_STX_ADR_ACK:            				// Own SLA+R has been received; ACK has been returned
      TWI_bufPtr = 0;                       		// Set buffer pointer to first data location
 
   case TWI_STX_DATA_ACK:           				// Data byte in TWDR has been transmitted; ACK has been received
      TWI_Mode = (1<<SLAVE | 1<<ACTIV | 1<<TRANSMIT);
																// Es werden solange Daten gesendet, bis der Master mit NACK beendet
																// damit kann der Pointer aus dem Array hinauslaufen. 
																// Eine berprfung wird eingebaut:
      TWDR = TWI_SLA_buf[TWI_bufPtr++];
		if (!(TWI_bufPtr < SLA_BUFFER_SIZE)) TWI_bufPtr = (SLA_BUFFER_SIZE - 1);
																// der Pointer wird nicht mehr incrementiert, wenn TWI_msgSize erreicht ist
      TWCR =(1<<TWEN)|                         	// TWI Interface enabled
            (1<<TWIE)|(1<<TWINT)|              	// Enable TWI Interupt and clear the flag to send byte
            (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   	// Send ACK
            (0<<TWWC);                         	//  
		break;

// case TWI_STX_DATA_ACK_LAST_BYTE: 				// Last data byte in TWDR has been transmitted (TWEA = 0); 
																// ACK has been received
																// da dem Slave nicht bekannt ist, wann der Master die bertragung
																// beenden wird, sollte dieser Status nicht auftreten,
																// statt dessen der nachfolgende:
																
   case TWI_STX_DATA_NACK:          				// Data byte in TWDR has been transmitted; NACK has been received. 
																// I.e. this could be the end of the transmission.
/*		if (TWI_SLA_bufPtr == TWI_msgSize) 			// Have we transceived all expected data?
			{
			TWI_statusReg.LastTransOK = TRUE;      // Set status bits to completed successfully. 
			}
		else                          				// Master has sent a NACK before all data where sent.
			{
			TWI_state = TWSR;                      // Store TWI State as errormessage.      
			}   */     
																// Im Slave-Transmitter Modus beendet der MAster mit einem NACK, 
																// definiert somit selbst das Ende 
																// ein Fehler liegt nur vor, wenn der Master mehr Bytes anfordert,
																// als bereitstehen - das kann der Slave dem Master aber nicht mitteilen ..
		TWI_statusReg.LastTransOK = TRUE;      	// Set status bits to completed successfully.
      TWI_Mode = 0;
                                                // Put TWI Transceiver in passive mode.
      TWCR = (1<<TWEN)|                         // Enable TWI-interface and release TWI pins
             (0<<TWIE)|(1<<TWINT)|              // Disable Interupt, release SCL
             (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Do not acknowledge on any new requests.
             (0<<TWWC);                         //
		break;     


// --- TWI Slave Receiver status codes	----------------------------------------
																
   case TWI_SRX_GEN_ACK_M_ARB_LOST: 				// Arbitration lost in SLA+R/W as Master; 
																// General call address has been received; ACK has been returned
		TWI_Arbitration 	= TWSR;						// Flag fr Arbitrierungsfehler setzen

   case TWI_SRX_GEN_ACK:            				// General call address has been received; ACK has been returned
      TWI_statusReg.GenAddressCall = TRUE;
																
																// ###############################################
   case TWI_SRX_ADR_ACK_M_ARB_LOST: 				// Arbitration lost in SLA+R/W as Master; 
																// own SLA+W has been received; ACK has been returned    
                                                // Dont need to clear TWI_statusRegister.generalAddressCall 
																// due to that it is the default state.
		TWI_Arbitration = TWSR;							// Flag fr Arbitrierungsfehler setzen

   case TWI_SRX_ADR_ACK: 								// Own SLA+W has been received ACK has been returned

      TWI_SLA_buf[0] = TWAR;							// fr durchgngigen Standard: Byte0 = TWI-Adresse
		TWI_SLA_bufPtr = 1;								// Set buffer pointer/counter to first data location
		TWI_Mode = (1<<SLAVE | 1<<ACTIV | 1<<RECEIVE);
                                                // Reset the TWI Interupt to wait for a new event.
      TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag to send byte
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Expect ACK on this transmission
             (0<<TWWC);                         //      
		break;

   case TWI_SRX_ADR_DATA_ACK:       				// Previously addressed with own SLA+W; data has been received; 
																// ACK has been returned
   case TWI_SRX_GEN_DATA_ACK:       				// Previously addressed with general call; 
																// data has been received; ACK has been returned
		if (TWI_SLA_bufPtr < SLA_BUFFER_SIZE)		// berprfung des Pointers
			{
			TWI_SLA_buf[TWI_SLA_bufPtr++] = TWDR;		
			TWI_statusReg.LastTransOK = TRUE;      // Set flag transmission successfull.       
                                                // Reset the TWI Interupt to wait for a new event.
			}
		// anstelle else: alle folgenden Bytes werden ignoriert (nicht gespeichert)
		// es tritt zwar wieder ein Fehler auf, aber der Master muss selbst dafr sorgen, dass er nicht
		// mehr Daten sendet, als der Slave empfangen kann
		TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag to send byte
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Send ACK after next reception
             (0<<TWWC);                         //  
      break;

   case TWI_SRX_STOP_RESTART:       				// A STOP condition or repeated START condition has been received 
																// while still addressed as Slave    
                                                // Put TWI Transceiver in passive mode.
      TWI_Mode = 0;
      TWCR = (1<<TWEN)|                         // Enable TWI-interface and release TWI pins
             (0<<TWIE)|(0<<TWINT)|              // Disable Interupt
             (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Do not acknowledge on any new requests.
             (0<<TWWC);                         //

		break;           
																// nicht behandelte Slave Receive Modi
// case TWI_SRX_ADR_DATA_NACK:      				// Previously addressed with own SLA+W; data has been received; 
																// NOT ACK has been returned
// case TWI_SRX_GEN_DATA_NACK:      				// Previously addressed with general call; 
																// data has been received; NOT ACK has been returned

// --- MASTER -----------------------------------------------------------------

// --- General TWI Master status codes	----------------------------------------
	case TWI_START:              						// START has been transmitted  
   case TWI_REP_START:          						// Repeated START has been transmitted
      TWI_bufPtr = 0;           						// Set buffer pointer to the TWI Address location
	
// --- TWI Master Transmitter status codes ------------------------------------
   case TWI_MTX_ADR_ACK:        						// SLA+W has been transmitted and ACK received
   case TWI_MTX_DATA_ACK:       						// Data byte has been transmitted and ACK received
      if (TWI_bufPtr < TWI_msgSize)
			{
			TWDR =TWI_MA_buf[TWI_bufPtr++];			// Adresse bzw. Daten nach TWDR schreiben
			TWI_Mode = (1<<MASTER | 1<<ACTIV |1<<TRANSMIT);
			TWCR =(1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to send byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|	// Send NACK after reception
               (0<<TWWC);
			}
      else                                      // Send STOP after last byte
			{
			if (TWI_MA_no_stop)							// wenn Restart gewnscht ist, dann kein STOP senden
				{
				TWI_MA_no_stop = 0;						// Flag zurcksetzen, sonst nichts tun - besonders kein STOP senden
				TWI_Mode = (1<<MASTER | 1<<ACTIV | 1<< TRANSMIT);
				TWCR = (0<<TWEN)|                   // TWI Interface disabled, PORTC wird wirksam
					(0<<TWIE)|(1<<TWINT)|            // Disable TWI Interrupt and clear the flag
					(0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Initiate NO STOP !! condition.
					(0<<TWWC);
				}

			else												// wenn TWI_no_stop = 0
				{
				TWI_statusReg.LastTransOK = TRUE;   // Set status bits to completed successfully.
				TWI_Mode = (1<<MASTER);
				TWCR = (1<<TWEN)|                   // TWI Interface enabled
               (0<<TWIE)|(1<<TWINT)|            // Disable TWI Interrupt and clear the flag
               (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)| // Initiate a STOP condition.
               (0<<TWWC);
				}
			}
			break;
		
// --- TWI Master Receiver status codes ---------------------------------------
   case TWI_MRX_DATA_ACK:                       // Data byte has been received and ACK transmitted
      TWI_MA_buf[TWI_bufPtr++] = TWDR;				// gelesene Daten in den Puffer schreiben

   case TWI_MRX_ADR_ACK:                        // SLA+R has been transmitted and ACK received
		TWI_Mode = (1<<MASTER | 1<<ACTIV | 1<<RECEIVE);
      if (TWI_bufPtr < (TWI_msgSize-1) )        // Detect the last byte to NACK it.
			{
			TWCR =(1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
               (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send ACK after reception
               (0<<TWWC);
			}
      else                                      // Send NACK after next reception
			{
			TWCR =(1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send NACK after reception
               (0<<TWWC);                          
			}    
		break; 

   case TWI_MRX_DATA_NACK:                      // Data byte has been received and NACK transmitted
      TWI_MA_buf[TWI_bufPtr] = TWDR;
      TWI_statusReg.LastTransOK = TRUE;         // Set status bits to completed successfully. 
		TWI_Mode = 0; 
		TWCR =(1<<TWEN)|                         	// TWI Interface enabled
            (0<<TWIE)|(1<<TWINT)|              	// Disable TWI Interrupt and clear the flag
            (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   	// Initiate a STOP condition.
            (0<<TWWC);
		break; 

// --- General TWI Master status codes ----------------------------------------    
   case TWI_ARB_LOST:                           // Arbitration lost
		TWI_Mode = (1<<MASTER | 1<<ACTIV ); 
		TWCR =(1<<TWEN)|                         	// TWI Interface enabled
            (1<<TWIE)|(1<<TWINT)|              	// Enable TWI Interupt and clear the flag
            (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|   	// Initiate a (RE)START condition.
            (0<<TWWC);                            
      break;

// case TWI_SRX_ADR_DATA_NACK:      				// Previously addressed with own SLA+W; data has been received; 
																// NOT ACK has been returned
// case TWI_SRX_GEN_DATA_NACK:      				// Previously addressed with general call; 
																// data has been received; NOT ACK has been returned
// case TWI_STX_DATA_ACK_LAST_BYTE: 				// Last data byte in TWDR has been transmitted (TWEA = 0); 
																// ACK has been received
// case TWI_BUS_ERROR:         						// Bus error due to an illegal START or STOP condition
// case TWI_MTX_ADR_NACK:                      	// SLA+W has been transmitted and NACK received
// case TWI_MRX_ADR_NACK:                      	// SLA+R has been transmitted and NACK received    
// case TWI_MTX_DATA_NACK:                     	// Data byte has been transmitted and NACK received
// case TWI_NO_STATE;                         	// No relevant state information available; TWINT = 0
// case TWI_BUS_ERROR:                        	// Bus error due to an illegal START or STOP condition

   default:     
      TWI_state = TWSR;                         // TWSR sichern, wenn Fehler vorliegen
      TWI_Mode = 0;                           	// Reset TWI Interface
      TWCR =(1<<TWEN)|                         	// Enable TWI-interface and release TWI pins
            (0<<TWIE)|(1<<TWINT)|              	// Disable Interupt   - nderung: clear Interrupt Flag
            (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   	// No Signal requests - nderung: Stop Condition senden
            (0<<TWWC);
				break;
	}
}

// --- [ eof ] ----------------------------------------------------------------